#include	<X11/Xlib.h>
#include        <X11/Xutil.h>
#include	<stdio.h>

#include        "vroot.h" /* for virtual root window */ 

#define NUMBER_OF_FRAGMENTS 40
#define SIZE_AT_LEAST 64
#define MAX_TRIAL 40

#define COLORS 1
#define COLORSTEP 16

typedef struct {
    Region region;		/* $BM?$($i$l$?%j!<%8%g%s(B */
    int imagex, imagey;		/* $B85%$%a!<%8$N>l=j(B */
    XRectangle clipbox;		/* $B%j!<%8%g%s$r4^$`;M3Q7A$NBg$-$5$H0LCV(B */
} FRAGMENT;

struct fragment_list_str {
    int id;
    FRAGMENT *fragment;            
    struct fragment_list_str *prev, *next;
    int dx, dy;
};
typedef struct fragment_list_str FRAGMENT_LIST;

#define NULL_LIST (FRAGMENT_LIST *) NULL

FRAGMENT_LIST *AllocFragmentList();

FRAGMENT * CreateFragment();
Region SetNewRegion();

#define OPT_ROOT "-root"
#define OPT_DISPLAY "-display"
#define OPT_DELAY "-delay"

#define MODE_ROOT 01

char *options[] = { OPT_ROOT, OPT_DISPLAY, OPT_DELAY, NULL };
enum options { OPT_ROOT_NUM, OPT_DISPLAY_NUM, OPT_DELAY_NUM};

Display		*dpy;
Window		win;
int		screen;
XGCValues	gcvals;
GC              copygc;
int             width, height;
Pixmap          screenpixmap;

Colormap        dcm;
unsigned long pmasks, pixels[COLORS];
XColor          xcolor[COLORS];

FRAGMENT_LIST *first;
FRAGMENT_LIST *last;
int id = 0;
int list_len;
int delay = 10000;

int mode = 0;

main(argc, argv)
int	argc;
char	**argv;
{
    char			*display = NULL;
    unsigned long		vmask;
    XSetWindowAttributes	xswat;

    int timeout, interval, preference, allowexp;
    int i = 1;
    int j;

    srandom(getpid());

    /* options */
    while( i!= argc ){
	j=0;
	while( options[j] != NULL ){
	    if( !strcmp( options[j], argv[i] ) ){
		switch( j ){
		 case OPT_ROOT_NUM:
		    mode |= MODE_ROOT;
		    break;
		 case OPT_DISPLAY_NUM:
		    i++;
		    if( i== argc ){
			fprintf( stderr, "-display need string.\n");
			error_exit();
		    }
		    display = argv[i];
		    break;
		 case OPT_DELAY_NUM:
		    i++;
		    if( i == argc ){
			fprintf( stderr, "-delay need integar.\n" );
			error_exit();
		    }
		    if(( delay = atoi( argv[i] ) ) == 0 ){
			fprintf( stderr, "-delay need non-zero integar.\n");
			error_exit();
		    }
		    break;
		 default:
		    fprintf( stderr, "unknown option %s\n", argv[i] );
		    error_exit();
		}
		i++;
		break;
	    } else j++;
	}
	if( options[j] == NULL ){ /* run over */
	  fprintf( stderr, "unknown option %s\n", argv[i] );
	  error_exit();
	}
    }

    if ((dpy = XOpenDisplay(display)) == NULL){
	fprintf(stderr, "can't open display\n");
	exit(0);
    }

    screen = DefaultScreen(dpy);
    dcm = DefaultColormap( dpy, screen );
#ifdef COLORANIM
    XAllocColorCells( dpy, dcm, False, &pmasks, 0, pixels, COLORS );
#endif

    if( mode & MODE_ROOT ){
    	win = RootWindow(dpy, screen); 
        /* for only xscreensaver */
	XSetWindowBackgroundPixmap (dpy, win, None);
	XUnmapWindow (dpy, win);
	XSync (dpy, True);
	sleep (5);		/* wait for everyone to swap in and handle exposes... */
	XMapWindow (dpy, win);
	XSync (dpy, True);
    } else {
	/* $B2hLLA4BN$KJ$$$Ho$5$k$h$&$J%&%$%s%I%&$r:n$k(B */
        xswat.override_redirect = True;
        vmask = CWOverrideRedirect;
	win = XCreateWindow(dpy, RootWindow(dpy, screen),
			    0, 0,
			    DisplayWidth(dpy, screen), 
			    DisplayHeight(dpy, screen),
			    0, CopyFromParent, InputOutput, CopyFromParent,
			    vmask, &xswat);
	XSelectInput( dpy, win,
		     ExposureMask | ButtonPressMask );
	XStoreName(dpy, win, argv[0]);
	XMapWindow(dpy, win);
    }
    
    /* $B2hLL$N>uBV$r5-21$9$k%T%C%/%9%^%C%W$r:n$k(B */
    {
	Window tmpw;
	int tmpi;
	XGetGeometry( dpy, win, &tmpw, &tmpi, &tmpi, 
		     &width, &height, &tmpi, &tmpi);
    }
    
    screenpixmap = XCreatePixmap( dpy, win, 
				 width, height, DefaultDepth( dpy, screen ));

    gcvals.graphics_exposures = False;
    gcvals.foreground = BlackPixel( dpy, screen );
    gcvals.line_width = 3;
    copygc = XCreateGC(dpy, win, GCForeground | GCLineWidth |
		       GCGraphicsExposures, &gcvals);
    
    while(1){
	XEvent e;
	if(mode & MODE_ROOT || 
	   XCheckWindowEvent( dpy, win, ExposureMask, &e )){
	    XCopyArea( dpy, win, screenpixmap, copygc, 
		      0, 0, width, height,
		      0, 0 );
	    break;
	}
    }
    do_all();
}

split_fragment( listp )
FRAGMENT_LIST *listp;
{
    Region cut, andr, xorr;
    XPoint points[4];
    XRectangle rect;
    int from, to;

    XClipBox( listp->fragment->region, &rect );
    
    if( rect.width < SIZE_AT_LEAST || rect.height < SIZE_AT_LEAST )
	return 0;

    points[0].x = points[3].x = rect.x; 
    points[0].y = points[1].y = rect.y;
    points[1].x = points[2].x = rect.x + rect.width;
    points[2].y = points[3].y = rect.y + rect.height;
    
    switch( random() % 2 ){
     case 0:
	if( rect.width < rect.height ){
	    points[0].y = random() % rect.height + rect.y;
	    points[1].y = random() % rect.height + rect.y;
	    from = 0; to = 1;
	} else {
	    points[1].x = random() % rect.width  + rect.x;
	    points[2].x = random() % rect.width  + rect.x;
	    from = 1; to = 2;
	}
	break;
     case 1:
	if( rect.width < rect.height ){
	    points[2].y = random() % rect.height + rect.y;
	    points[3].y = random() % rect.height + rect.y;
	    from = 2; to = 3;
	} else {
	    points[0].x = random() % rect.width  + rect.x;
	    points[3].x = random() % rect.width  + rect.x;
	    from = 0; to = 3;
	}
	break;
    }
    
    XSetRegion( dpy, copygc, listp->fragment->region );
    XDrawLine( dpy, screenpixmap, copygc, 
	      points[from].x, points[from].y,
	      points[to].x, points[to].y );

    cut = XPolygonRegion( points, 4, EvenOddRule );
    andr = XCreateRegion();
    xorr = XCreateRegion();
    XIntersectRegion( listp->fragment->region, cut, andr );
    XXorRegion( listp->fragment->region, andr, xorr );
    
    if( XEmptyRegion( andr ) || XEmptyRegion( xorr )) {
	XDestroyRegion( cut );
	XDestroyRegion( andr );
	XDestroyRegion( xorr );
	return 0;
    }
    AddNewFragment( andr );
    AddNewFragment( xorr );
    XDestroyRegion( cut );
    EraseFragmentList( listp );
    return 1;
}

FRAGMENT_LIST *AllocFragmentList()
{
    FRAGMENT_LIST *listp;

    if( (listp = (FRAGMENT_LIST *)malloc(sizeof( FRAGMENT_LIST ))) 
       == NULL_LIST){
	fprintf(stderr, "No memory for a fragment list\n");
	exit(1);
    }
    listp->id = id++;
    return listp;
}

AddNewFragment( reg )
Region reg;
{
    FRAGMENT_LIST *listp;

    listp = AllocFragmentList();
    listp->fragment = CreateFragment( reg );
    listp->prev = last;
    if( last->next != NULL_LIST )
	fprintf( stderr, "OOps?? last.\n");
    last->next = listp;
    last = listp;
    last->next = NULL_LIST;
    list_len ++;
    listp->dx = random() % 9 - 4;
    listp->dy = random() % 9 - 4;
}

EraseFragmentList( listp )
FRAGMENT_LIST *listp;
{
    if ( listp->prev == NULL_LIST )
	first = listp->next;
    else
	listp->prev->next = listp->next;
    
    if( listp->next == NULL_LIST )
	last = listp->prev;
    else
	listp->next->prev = listp->prev;

    FreeFragment( listp->fragment );
    free( listp );
    list_len--;
}

do_all()
{
    XPoint points[4];
    Region reg;
    FRAGMENT *frag;
    GC putgc;
    int i,j;
    FRAGMENT_LIST *listp;
    unsigned int red, blue, green;
    unsigned int gred, gblue, ggreen;
    int step;
    
    points[0].x = points[3].x = 0; 
    points[0].y = points[1].y = 0;
    points[1].x = points[2].x = width;
    points[2].y = points[3].y = height;
    
    first = last = AllocFragmentList();
    first->fragment = CreateFragment(XPolygonRegion( points, 4, EvenOddRule ));
    first->prev = first->next = NULL_LIST;
    list_len = 1;

    listp;
    for(i=0; i<NUMBER_OF_FRAGMENTS; i++) {
	int trial;
	trial = 0;
	do{
	    listp = first;
	    if( list_len > 1 )
		for(j=0; j< random()%(list_len - 1); j++)
		    listp = listp->next;
	    trial ++;
	    if( trial > MAX_TRIAL ) break;
	} while (!split_fragment( listp ));
    }
    
    putgc = XCreateGC(dpy, win, GCGraphicsExposures, &gcvals);
    XSetForeground( dpy, putgc, pixels[0] );
    reg = XPolygonRegion( points, 3, EvenOddRule );
    frag = CreateFragment( reg );
    
    step = 0;
    gred = 0; gblue = 0; ggreen = 0;
    
    while(1){
	XEvent e;
	if( !(mode & MODE_ROOT) && 
	   XCheckWindowEvent( dpy, win, ButtonPressMask, &e)) 
	    exit (0);
#ifdef COLORAMIM
	if( step == 0 ){
	    red = gred;
	    blue = gblue;
	    green = ggreen;
	    gred = random()%64;
	    gblue = random()%64;
	    ggreen = random()%64;
	    step = COLORSTEP;
	}
	xcolor[0].pixel = pixels[0];
	xcolor[0].red = (red * step + gred * (COLORSTEP - step))/COLORSTEP*1000;
	xcolor[0].blue = (blue * step + gblue * (COLORSTEP - step))/COLORSTEP*1000;
	xcolor[0].green = (green * step + ggreen * (COLORSTEP - step))/COLORSTEP*1000;
	xcolor[0].flags = DoRed | DoGreen | DoBlue;
	XStoreColors( dpy, dcm, xcolor, COLORS );
	step --;
#endif

	listp = first;
	do{
	    Region prevregion;
	    
	    prevregion = XCreateRegion();
	    prevregion = SetNewRegion( listp, prevregion );
	    
	    MoveFragment( listp->fragment, listp->dx, listp->dy );
	    
	    /* out check */
	    if( XRectInRegion( listp->fragment->region, 0, 0, width, height )
	       == RectangleOut ){
		MoveFragment( listp->fragment, -listp->dx, -listp->dy );
		listp->dx = random() % 9 - 4;
		listp->dy = random() % 9 - 4;
	    }

	    ClearFragmentOfList( dpy, win, putgc, listp, prevregion );
	    PutFragmentOfList( dpy, screenpixmap, win, putgc, listp );
	    listp = listp->next;
	    
	    XDestroyRegion( prevregion );
	}while( listp != NULL_LIST );
	XFlush(dpy);
	usleep(delay);
    }
}

FRAGMENT * CreateFragment( region )
Region region;
{
    FRAGMENT *frag;
    
    if((frag = (FRAGMENT *) malloc( sizeof( FRAGMENT ))) == (FRAGMENT *)NULL ){
	fprintf(stderr, "No memory for a fragment.\n");
	exit(1);
    }
    
    frag->region = region;
    XClipBox( region, &frag->clipbox );
    frag->imagex = frag->clipbox.x;
    frag->imagey = frag->clipbox.y;

    return frag;
}

FreeFragment( frag )
FRAGMENT *frag;
{
    XDestroyRegion( frag->region );
    free( frag );
}

Region SetNewRegion( listp, clip )
FRAGMENT_LIST *listp;
Region clip;
{
    FRAGMENT_LIST *itep;

    /* get a copy */
    XUnionRegion( listp->fragment->region, listp->fragment->region, clip ); 
    itep = listp->next;
    while( itep != NULL_LIST ){
	XSubtractRegion( clip, itep->fragment->region, clip );
	itep = itep->next;
    }
    return clip;
}

PutFragmentOfList( disp, src, dst, gc, listp )
Display *disp;
Drawable src, dst;
GC gc;
FRAGMENT_LIST *listp;
{
    Region clip;
    
    clip = XCreateRegion();
    clip = SetNewRegion( listp, clip );
    
    XSetRegion( disp, gc, clip );
    XCopyArea( disp, src, dst, gc,
	      listp->fragment->imagex, listp->fragment->imagey,
	      listp->fragment->clipbox.width, listp->fragment->clipbox.height,
	      listp->fragment->clipbox.x, listp->fragment->clipbox.y);
    XDestroyRegion( clip );
}

ClearFragmentOfList( disp, dst, gc, listp, prevregion )
Display *disp;
Drawable dst;
GC gc;
FRAGMENT_LIST *listp;
Region prevregion;
{
    Region clip;
    XRectangle rect;
    
    clip = XCreateRegion();
    clip = SetNewRegion( listp, clip );
    
    /* $B$1$9$H$3$m$+$i$D$.$+$/$H$3$m$r$N$>$/(B*/
    XSubtractRegion( prevregion, clip, clip );
    
    XSetRegion( disp, gc, clip );
    XClipBox( clip, &rect );

    XFillRectangles( disp, dst, gc, &rect, 1);
    XDestroyRegion( clip );
}

PutFragment( disp, src, dst, gc, frag )
Display *disp;
Drawable src, dst;
GC gc;
FRAGMENT *frag;
{
    XSetRegion( disp, gc, frag->region );
    XCopyArea( disp, src, dst, gc,
	      frag->imagex,        frag->imagey,
	      frag->clipbox.width, frag->clipbox.height,
	      frag->clipbox.x,     frag->clipbox.y);
}

ClearFragment( disp, dst, gc, frag )
Display *disp;
Drawable dst;
GC gc;
FRAGMENT *frag;
{
    XSetRegion( disp, gc, frag->region );
    XFillRectangles( disp, dst, gc, &frag->clipbox, 1);
}

MoveFragment( frag, dx, dy )
FRAGMENT *frag;
int dx, dy;
{
    XOffsetRegion( frag->region, dx, dy );
    frag->clipbox.x += dx;
    frag->clipbox.y += dy;
}

error_exit()
{
    fprintf(stderr, "xcrushroot [options]\n" );
    fprintf(stderr, "            -root (FOR xscreensaver ONLY !!!)\n");
    fprintf(stderr, "            -display DISPLAY_NAME\n");
    fprintf(stderr, "            -delay DELAY(10000 as default)\n");
    fprintf(stderr, "by J.P.Furuse\n");
    exit(1);
}

